package uoc.ei.tads;
/**
 * Interfaz que define las operaciones de cualquiera arbol (tree).
 *
 * Los rboles son estructuras que relacionan sus elementos, llamados
 * nodos, formando jerarquas: todo nodo (excepto la raz que es la cabeza de
 * la jerarqua) es descendiente de un nodo nico, y puede ser ascendente
 * de otros nodos (cuando no tiene descendientes se llama hoja). Cuando un nodo
 * puede tener un nmero indeterminado de hijos hablamos de rboles generales
 * (general tree) y, si tiene un nmero fijo N, de rboles de orden N (n-ary
 * tree); en estos ltimos destaca el caso de N = 2, los llamados rboles
 * binarios (binary tree).
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informacin,
 *          Universitat Oberta de Catalunya (UOC)
 * @version 2.0.0
 */
public interface Arbol<E>  extends Contenedor<E>
{
   /**
    * Accesor de lectura de la raz del arbol, si hay.
    * @return  raz del arbol
    */
   public Posicion<E> raiz();

   /**
    * Mtodo que soporta mltiples recorridos, de las posiciones hijas
    * de la posicin de referencia, simultneos e independientes entre ellos.
    * @param padre  posicin de referencia
    * @return  enumeracin de las posiciones hijas
    */   public Recorrido<E> hijos(Posicion<E> padre);

    /**
     * Comprueba si el arbol o subrbol tiene algn hijo.
     * @param pos  posicin de referencia
     * @return  falso o cierto, segn si tiene algn hijo o no tiene ningn
     */
   public boolean esHoja(Posicion<E> pos);

   /**
    * Aade un elemento como nuevo hijo de la posicin recibida, si se puede.
    * Si el padre es null, se aade en la raz; si es una hoja, se aade
    * como primero hijo (hijo ms a la izquierda).
    * @param padre  posicin de referencia
    * @param elem  elemento que se quiere aadir al arbol
    * @return  nuevo hijo; o la raz, si el padre era null
    */
   public Posicion<E> insertar(Posicion<E> padre,E elem);

   /**
    * Reemplaza el elemento contenido en la posicin recibida.
    * @param elem  nuevo elemento
    * @param pos  posicin de referencia
    * @return  elemento que haba a la posicin
    */
   public E reemplazar(Posicion<E> pos, E elem);

   /**
    * Intercambia en el arbol los elementos contenidos a las
    * posiciones recibidas. El arbol resultante contendr los mismos
    * elementos que tena, excepto por los dos elementos recibos como
    * parmetro, que aparecern intercambiados.<p>
    * El hecho de si los objetos posicin se intercambian, o bien el que
    * se intercambia son los elementos contenidos en las posiciones, depender
    * de cada implementacin concreta de arbol.
    * @param pos1  primera de las dos posiciones de referencia
    * @param pos2  segunda de las dos posiciones de referencia
    */
   public void intercambiar(Posicion<E> pos1, Posicion<E> pos2);

   /**
    * Borra el subrbol representado por la posicin hijo, si se puede.
    * Si la posicin del padre es null, borra el arbol entero.
    * @param padre  posicin del padre; puede ser null
    * @param hijo  posicin del hijo
    */
   public void borrar(Posicion<E> padre, Posicion<E> hijo);

   /** Recorrido de las posiciones del arbol.
    * @return  enumeracin de las posiciones del contenedor por nivel
    */
   public Recorrido<E> posiciones();

   /** Recorrido en preorden de las posiciones del arbol.
    * @return  enumeracin de las posiciones del contenedor preorden
    */
   public Recorrido<E> recorridoPreorden();

   /** Recorrido en posorden de las posiciones del arbol.
    * @return  enumeracin de las posiciones del contenedor posorden
    */
   public Recorrido<E> recorridoPostorden();


   /** Recorrido por niveles de las posiciones del arbol.
    * @return  enumeracin de las posiciones del contenedor por niveles
    */
   public Recorrido<E> recorridoPorNiveles();
}
