package uoc.ei.tads;
/**
 * Clase que implementa las operaciones de cualquier arbol binario, el cual
 * se caracteriza por organizar sus elementos (nodos) formando una
 * jerarqua: todo nodo (excepto la raz) es
 * descendiente de un nodo nico, y puede ser ascendente de un mximo de dos nodos
 * (cuando no tiene descendientes se nombra a hoja).
 *
 * En la representacin secuencial del arbol se organizan los nodos dentro de
 * de un vector de manera que los hijos y el padre son accesibles aplicando
 * solo una formula. La primera posicin del vector contendr siempre
 * la raz del arbol; si un nodo est en la posicin [k], su hijo
 * izquierdo estar en la [2*k] y su hijo derecho en la [2*k+1].
 *
 * Esta representacin tiene dos operaciones modificadoras (aadir al final
 * y borrar la ltima posicin) que permiten mantener una estructura
 * de arbol quasicompleto; es decir, no habr ningna posicin libre en un
 * recorrido por niveles del arbol. Dado que la primera posicin [0] del
 * vector no se utiliza para facilitar los clculos mencionados, si el nmero
 * de casillas del vector es potencia de dos, el arbol podr llegar a ser
 * completo.
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informacin,
 *          Universitat Oberta de Catalunya (UOC)
 * @version 2.0.0
 */
public class ArbolBinarioVectorImpl<E>  extends ArbolBinario<E>
{
   /** Atributo que en Java 1.5 determina la compatibilidad entre objetos
    * serialitzables de la misma clase. El identificador se calcula
    * mediante un mtodo de la clase Utilidades. */
   private static final long serialVersionUID = Utilidades.getSerialVersionUID();
	
	
   /** Capacidad mxima, por defecto, del contenedor. */
   public static final int MAXIMO_ELEMENTOS_POR_DEFECTO = 256;

   /** Nmero de elementos que hay actualmente al contenedor. */
   protected int n;

   /** Tabla de elementos del contenedor. Las posiciones empiezan por el cero.*/
   protected Rango<E>[] elems;

   /** Constructor sin parmetros (capacidad mxima, por defecto). */
   public ArbolBinarioVectorImpl() {
   	  this(MAXIMO_ELEMENTOS_POR_DEFECTO);
   }

   /**
    * Constructor con un parmetro. Crea un vector con una capacidad de
    * max.
    * @exception ExcepcionParametroIncorrecto  si la capacidad mxima del
    *                                         nuevo arbol es negativa
    * @pre max>=0 , new ExcepcionParametroIncorrecto("el nmero de elementos mximo no puede ser negativo")
    * @param max  nmero mximo de elementos del arbol 
    */
   public ArbolBinarioVectorImpl(int max) {
      elems = new Rango[max];
      n = 0;
   }

   /**
    * Accesor de lectura del nmero de elementos que hay al contenedor.
    * @return  nmero de elementos que contiene actualmente
    */
   public int numElems() { return  n; }

   /**
    * Mtodo para comprobar si el contenedor est vaco.
    * @return  cierto o falso, segn si est vaco o no lo est
    */
   public boolean estaVacio() { return  ( n == 0 ); }

   /**
    * Mtodo para comprobar si el contenedor est lleno.
    * @return  cierto o falso, segn si est lleno o no lo est
    */
   public boolean estaLleno() { return  ( n == elems.length ); }

   /**
    * Accesor de lectura de la raz del arbol, si hay.
    * @pre !estaVacio(), new ExcepcionContenedorVacio("No se puede acceder a la raz de un arbol vaco")
    * @exception ExcepcionContenedorVacio  si el arbol est vaco
    * @return  raz del arbol
    */
   public Posicion<E> raiz() { return  elems[0]; }
   
   
   /**
    * Aade un elemento como nuevo hijo de la posicin escogida.
    * Si el padre es null, se aade en la raz; si es una hoja, se aade
    * como hijo izquierdo; y si el padre tiene nicamente un hijo, se aade como
    * el otro hijo.
    * @param padre  posicin de referencia
    * @param elem  elemento que se quiere aadir al arbol
    * @pre posicionValida(padre) &&
    * 	    (indiceHijoIzquierdo(padre)<maximo && !tieneHijoIzquierdo(padre)) ||
    * 	    (indiceHijoDerecho(padre)<maximo && !tieneHijoDerecho(padre)), ExcepcionPosicionInvalida
    * @exception ExcepcionPosicionInvalida  Si la posicin de padre es invlida para poder
    * aadir un hijo. Eso incluye tambin los casos en que el padre ya tenga dos hijos y
    * el caso en que los hijos caigan fuera del rango mximo permitido al crear el arbol.
    * @return La posicin del nodo aadido.
    */
   public Posicion<E> insertar(Posicion<E> padre, E elem)
   {
   	  Posicion<E> posicionNuevoHijo;
   	  if (hijoIzquierdo(padre)!=null)
   	  	 posicionNuevoHijo=insertarHijoDerecho(padre,elem);
   	  else
   	  	 posicionNuevoHijo=insertarHijoIzquierdo(padre,elem);
   	  return posicionNuevoHijo;
   }
   

   /**
    * Aade un elemento como hijo izquierdo de la posicin escogida.
    * @param padre  posicin de referencia
    * @param elem  elemento que se quiere aadir al arbol
    * @pre posicioValida(padre) &&
    * 	    indiceHijoIzquierdo(padre)<maxim && !tieneHijoIzquierdo(padre)
    * @exception ExcepcionPosicionInvalida  Si la posicin de padre es invlida para poder
    * aadir un hijo. Eso incluye tambin los casos en que el padre ya tenga hijo izquierdo
    * y el caso en que este caiga fuera del rango mximo permitido al crear el arbol.
    * @return La posicin del nodo aadido.
    */
   public Posicion<E> insertarHijoIzquierdo(Posicion<E> padre, E elem)
   {
   	  int indice = getIndiceHijoIzquierdo(padre);
   	  n++;
   	  elems[indice]=new Rango<E>(indice,elem);
   	  return elems[indice];
   }
   

   /**
    * Aade un elemento como hijo derecho de la posicin escogida, si se puede.
    * @param padre  posicin de referencia
    * @param elem  elemento que se quiere aadir al arbol
    * @pre posicionValida(padre) &&
    * 	    indiceHijoDerecho(padre)<maximo && !tieneHijoDerecho(padre)
   * @exception ExcepcionPosicionInvalida  Si la posicin de padre es invlida para poder
    * aadir un hijo. Eso incluye tambin los casos en que el padre ya tenga hijo derecho
    * y el caso en que este caiga fuera del rango mximo permitido al crear el arbol.
    * @return La posicin del nodo aadido.
    */
   public Posicion<E> insertarHijoDerecho(Posicion<E> padre, E elem)
   {
 	  int indice = getIndiceHijoDerecho(padre);
   	  n++;
   	  elems[indice]=new Rango<E>(indice,elem);
   	  return elems[indice];
   }

   
   /**
    * Accesor de lectura del padre de una posicin del arbol, si hay.
    * @param pos  posicin de referencia
    * @return  padre de la posicin recibida; o null si era la raz.
    */
   public Posicion<E> padre(Posicion<E> pos)
   {
   	  if (getIndice(pos)==0) return null;
   	  return elems[getIndicePadre(pos)];
   }

   /**
    * Accesor de lectura del hijo izquierdo de una posicin del arbol.
    * @param pos  posicin de referencia
    * @pre posicionValida(pos)
    * @exception ExcepcionPosicionInvalida  si la posicin no es vlida.
    * @return  hijo izquierdo de la posicin escogida; o null si no tiene.
    */
   public Posicion<E> hijoIzquierdo(Posicion<E> pos)
   {
 	  int indice=getIndiceHijoIzquierdo((Rango<E>)pos);
   	  if (indice>=elems.length || elems[indice]==null)
   	  	 return null;
   	  return elems[indice];
   }

   /**
    * Accesor de lectura del hijo derecho de una posicin del arbol.
    * @param pos  posicin de referencia
    * @pre posicionValida(pos)
    * @exception ExcepcionPosicionInvalida  si la posicin es null o no
    *            vlida
    * @return  hijo derecho de la posicin escogida; o null si no tiene.
    */
   public Posicion<E> hijoDerecho(Posicion<E> pos)
   {
	  int indice=getIndiceHijoDerecho((Rango<E>)pos);
   	  if (indice>=elems.length || elems[indice]==null)
   	  	 return null;
   	  return elems[indice];
   }

   /**
    * Borra el subrbol representado por la posicin hijo, si se puede.
    * Si la posicin del padre es null, borra el arbol entero.
    * @param padre  posicin del padre; puede ser null
    * @param hijo  posicin del hijo
    * @pre esPadreDe(padre,hijo)
    * @exception ExcepcionPosicionInvalida  si alguna posicin es no vlida o
    * bien padre no es el padre de hijo.
    */
   public void borrar(Posicion<E> padre, Posicion<E> hijo)
   {
 	  Posicion<E> nietoIzquierdo = hijoIzquierdo(hijo);
   	  Posicion<E> nietoDerecho = hijoDerecho(hijo);
   	  if (nietoIzquierdo!=null)
   	  	  borrar(hijo,nietoIzquierdo);
   	  if (nietoDerecho!=null)
   	  	  borrar(hijo,nietoDerecho);
   	  int indiceHijo=getIndice(hijo);
   	  n--;
   	  elems[indiceHijo]=null;
   }


   /**
    * Reemplaza el elemento contenido en la posicin escogida.
    * @param elem  nuevo elemento
    * @param pos  posicin de referencia
    * @pre posicioValida(pos)
    * @return  elemento que haba a la posicin antes de reemplazarlo.
    */
   public E reemplazar(Posicion<E> pos, E elem)
   {
   	  Rango<E> rango=(Rango<E>)pos;
   	  E old=rango.getElem();
   	  rango.setElem(elem);
   	  return old;
   }

   /**
    * Intercambia dos posiciones dentro de de l'arbol. Los objetos
    * posicin se intercambian de posicin, con lo cual su
    * contenido se mantiene constante.
    * @param pos1  primera de las dos posiciones de referencia.
    * @param pos2  segunda de las dos posiciones de referencia.
    * @pre posicionValida(pos1) && posicionValida(pos2)
    */
   public void intercambiar(Posicion pos1, Posicion pos2)
   {
 	  Rango<E> r1=(Rango<E>)pos1;
   	  Rango<E> r2=(Rango<E>)pos2;
   	  int indice1=r1.getIndice();
   	  int indice2=r2.getIndice();
   	  r1.setIndice(indice2);
   	  r2.setIndice(indice1);
   	  elems[indice1]=r2;
   	  elems[indice2]=r1;
   }


   private int getIndice(Posicion<E> pos) {
   	  if (pos==null) return -1;
	  return ((Rango)pos).getIndice();
   }


   private int getIndicePadre(Posicion<E> pos) {
   	  int indice=((Rango)pos).getIndice();
   	  if (indice==0) return -1;
	  return (indice-1)/2;
   }


   private int getIndiceHijoIzquierdo(Posicion<E> pos) {
 	  if (pos==null) return 0;
	  return ((Rango)pos).getIndice()*2+1;
   }


   private int getIndiceHijoDerecho(Posicion<E> pos) {
 	  if (pos==null) return 0;
 	  return ((Rango)pos).getIndice()*2+2;
 }
 

}
