package uoc.ei.tads;
/**
 * Les arboles son estructuras que relacionan sus elementos, llamados
 * nodos, formando jerarquas: todo nodo (excepto la raz) es descendiente
 * de un nodo nico, y puede ser ascendente
 * de otros nodos (cuando no tiene descendientes se nombra a hoja). Cuando un nodo
 * puede tener un nmero indeterminado de hijos hablamos de rboles generales
 * (general tree) y, si tiene un nmero fijo N, de rboles de orden N (n-ary
 * tree); en estos ltimos destaca el caso de N = 2, les llamados rboles
 * binarios (binary tree).
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informacin,
 *          Universitat Oberta de Catalunya (UOC)
 * @version 2.0.0
 */
public class ArbolGeneralDelegImpl<E>  extends ArbolAbstracto<E> {
   /** Atributo que en Java 1.5 determina la compatibilidad entre objetos
    * serialitzables de la misma clase. El identificador se calcula
    * mediante un mtodo de la clase Utilidades. */
   private static final long serialVersionUID = Utilidades.getSerialVersionUID();
	
	
   /** Arbol binario que implementa las operaciones por delegacin. */
   protected ArbolBinarioEncadenadoImpl<E> ab;
   
   
   public ArbolGeneralDelegImpl() {
	   ab = crearArbolRepresentacion();
   }
   
   
   protected ArbolBinarioEncadenadoImpl<E> crearArbolRepresentacion() {
	   return new ArbolBinarioEncadenadoImpl<E>();
   }


   /**
    * Accesor de lectura del nmero de elementos que hay en el contenedor.
    * @return  nmero de elementos que contiene actualmente
    */
   public int numElems() { return  ab.numElems(); }


   /**
    * Accesor de lectura de la raz del arbol, si hay.
    * @return  raz del arbol
    */
   public Posicion<E> raiz() { return  ab.raiz(); }

   /**
    * Mtodo que soporta mltiples recorridos, de las posiciones hijas
    * de la posicin de referencia, simultneos e independientes entre ellos.
    * @param padre  posicin de referencia
    * @exception ExcepcionPosicionInvalida  si la posicin es null o no
    *            vlida
    * @return  enumeracin de las posiciones hijas
    */
   public Recorrido<E> hijos(Posicion<E> padre)
   {
      if (estaVacio() || (padre == null))
         throw  new ExcepcionPosicionInvalida();

      return new RecorridoHijos<E>(this,padre);
   }

   /**
    * Aade un elemento como nuevo hijo de la posici recibida, si se puede.
    * Si el padre es null, se aade en la raz; si es una hoja, se aade
    * como primer hijo (hijo ms a la izquierda);  sino, se aade a
    * la primera posicin libre (de izquierda a derecha).
    * @param padre  posicin de referencia
    * @param elem  elemento que se quiere aadir al arbol
    * @return  nuevo hijo; o la raz, si el padre era null
    */
   public Posicion<E> insertar(Posicion<E> padre, E elem)
   {
      if ( (padre == null) || esHoja(padre) )
         return  ab.insertar(padre, elem);

      padre = ab.hijoIzquierdo(padre);
      while (ab.hijoDerecho(padre) != null)  padre = ab.hijoDerecho(padre);

      return  ab.insertarHijoDerecho(padre, elem);
   }

   /**
    * Reemplaza el elemento contenido en la posicin rebuda.
    * @param elem  nuevo elemento
    * @param pos  posicin de referencia
    * @return  elemento que haba en la posicin
    */
   public E reemplazar(Posicion<E> pos, E elem)
   {
      return  ab.reemplazar(pos, elem);
   }

   /**
    * Intercambia los elementos contenidos en las las posiciones recibidas.
    * Las posiciones no son intercambiadas.
    * @param pos1  primera de las dos posiciones de referencia
    * @param pos2  segunda de las dos posiciones de referencia
    */
   public void intercambiar(Posicion<E> pos1, Posicion<E> pos2)
   {
      ab.intercambiar(pos1, pos2);
   }

   /**
    * Borra el subrbol representado por la posicin hijo, si se puede.
    * Si la posicin del padre es null, borra el arbol entero.
    * @param padre  posicin del padre; puede ser null
    * @param hijo  posicin del hijo
    * @exception ExcepcionPosicionInvalida  si alguna posicin es no vlida
    */
   public void borrar(Posicion<E> padre, Posicion<E> hijo)
               throws  ExcepcionPosicionInvalida
   {
      if (estaVacio() || (hijo == null))
         throw  new ExcepcionPosicionInvalida();

      if (padre == null)  ab.borrar(padre, hijo);
      else
      {
         if (hijo != ab.hijoIzquierdo(padre))
         {
            for ( padre = ab.hijoIzquierdo(padre);
                  ab.hijoDerecho(padre) != hijo;
                  padre = ab.hijoDerecho(padre) );
         }
         ab.reemplazarSubarbol(padre, hijo, ab.hijoDerecho(hijo));
      }
   }

   /**
    * Clase que proporciona un recorrido de las posiciones. Basada en el
    * pattern Iterator, soporta mltiples recorridos simultneos y
    * independientes del contenedor. Es sensible a eventuales alteraciones de
    * la estructura de posiciones.
    * @see  Recorrido#haySiguiente()
    * @see  Recorrido#siguiente()
    */
   protected static class RecorridoHijos<E> implements Recorrido<E>
   {

      private static final long serialVersionUID = Utilidades.getSerialVersionUID();

      
      /** Arbol que se est recorriendo. */
      protected ArbolGeneralDelegImpl<E> arbol;

      
      /** Posicin auxiliar. */
      protected Posicion<E> hijo;


      /**
       * Constructor.
       * @param padre  posicio de referencia; se espera que sea una
       *              posicin vlida
       */
      public RecorridoHijos(ArbolGeneralDelegImpl<E> arbol,Posicion<E> padre) {
      	 this.arbol=arbol;
         hijo = arbol.ab.hijoIzquierdo(padre);
      }

      
      /**
       * Comprueba si hay una primera o siguiente posicin. Es sensible a
       * eventuales alteraciones de la estructura de posiciones. Retorna falso
       * si est vaco o ya se ha visitado la ltima posicin.
       * @return  cierto o falso, segn si se puede avanzar o no se puede
       */
      public boolean haySiguiente() { return  hijo != null; }

      
      /**
       * Primero avanza, si se puede, y despus retorna la posicin.
       * Si no hay siguiente posicin lanza una excepcin.
       * @return  siguiente posicin
       */
      public Posicion<E> siguiente()  throws  ExcepcionPosicionInvalida {
          if ( ! haySiguiente() )  throw new ExcepcionPosicionInvalida(
                                           "No hay siguiente");

         Posicion<E> temp = hijo;
         hijo = arbol.ab.hijoDerecho(hijo);
         return  temp;
      }
   }
}
