package uoc.ei.tads;
/**
 * Clase que empareja dos objetos. El primero (clave) debe identificar
 * el par, el segundo elemento (valor) almacena una informacin
 * asociada a la clave y puede ser cualquier objeto.
 *
 * Implementa la interfaz Serializable para poder convertir a cadenas o
 * flujos de bytes (streams) los objetos del contenedor y grabarlos o
 * transmitirlos.
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informacin,
 *          Universitat Oberta de Catalunya (UOC)
 * @version 2.0.0
 */
public class ClaveValor<C,E>  implements java.io.Serializable, Comparable<ClaveValor<C,E>>
{
   /** Atributo que en Java 1.5 determina la compatibilidad entre objetos
    * serialitzables de la misma clase. El identificador se calcula
    * mediante un mtodo de la clase Utilidades. */
   private static final long serialVersionUID = Utilidades.getSerialVersionUID();
	
	
   /** Clave o ndice que permite identificar la par. */
   protected C clave;

   /** Informacin asociada a la clave. Puede tener valor nulo. */
   protected E valor;

   /**
    * Constructor con dos parmetros.
    * @param clave  elemento comparable 
    * @param valor  informacin que debe almacenar el objeto
    * @pre clave!=null, new ExcepcionParametroIncorrecto("la clave debe ser diferente de null")
    * @exception ExcepcionParametroIncorrecto  si la clave es null
    */
   public ClaveValor(C clave, E valor)
   {
      this.clave = clave;
      this.valor = valor;
   }

   /**
    * Accesor de lectura de la clave o ndice.
    * @return  valor de la informacin asociada a la clave
    */
   public C getClave() { return  clave; }

   /**
    * Accesor de escritura de la informacin asociada a la clave.
    * @param valor  nuevo valor de la informacin almacenada
    */
   public void setValor(E valor) { this.valor = valor; }

   /**
    * Accesor de lectura de la informacin asociada a la clave.
    * @return  valor de la informacin almacenada
    */
   public E getValor() { return  valor; }

   /**
    * Mtodo que sobrescribe Object.hashCode(). Delega en el mismo mtodo
    * implementado o heredado por la clave.
    * @return  cdigo de dispersin correspondiente a la clave
    */
   public int hashCode() { return  clave.hashCode(); }

   /**
    * Mtodo que sobrescribe Object.equals(Object obj). Delega en el
    * mismo mtodo implementado o heredado por la clave.
    * @return  cierto si el objeto recibo como parmetro es de la clase
    *          ClaveValor y tiene la misma clave; sino retorna falso
    */
   public boolean equals(Object clauValor)
   {
   	  if (clauValor==null) return false;
   	  return clave.equals( ((ClaveValor)clauValor).getClave() );
   }

   
   /**
    * Mtodo que implementa Comparable.compareTo(Object o). Delega en el
    * mismo mtodo implementado o heredado por la clave.
    * @param claveValor  objeto de la misma clase (ClaveValor)
    * @return  un entero negativo, cero o positivo, segn si la clave
    *          de este objeto (this.clave) es ms pequea, igual o ms
    *          grande que la clave del objeto de la misma clase recibo como
    *          parmetro (claveValor.clave)
    */
   public int compareTo(ClaveValor<C,E> claveValor)
   {
      return  ((Comparable<C>)clave).compareTo(
                 (claveValor).getClave() );
   }

   
   /**
    * Mtodo que define la conversin del objeto a String por
    * facilitar la depuracin del cdigo.
    * @return  cadena de caracteres con la clave y el valor
    */
   public String toString() { return  ( "[" + clave + ": " + valor +"]"); }
}
