package uoc.ei.tads;
/**
 * Clase que implementa las operaciones bsicas de un conjunto por
 * delegacin en una tabla de dispersin.
 *
 * Los conjuntos son estructuras que almacenan elementos no repetidos. La
 * clase de los objetos debe disponer de una operacin de igualdad.
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informacin,
 *          Universitat Oberta de Catalunya (UOC)
 * @version 2.0.0
 */
public class ConjuntoTablaImpl<E>  extends ConjuntoAbstracto<E>
{
   /** Atributo que en Java 1.5 determina la compatibilidad entre objetos
    * serialitzables de la misma clase. El identificador se calcula
    * mediante un mtodo de la clase Utilidades. */
   private static final long serialVersionUID = Utilidades.getSerialVersionUID();


   /** Tabla de dispersin que implementa las operaciones por delegacin. */
   protected TablaDispersion<E,E> td;

   
   /** Constructor sin parmetros (medida de la tabla por defecto). */
   public ConjuntoTablaImpl()
   {
   	td = new TablaDispersion<E,E>();
   }

   
   /** Constructor con un parmetro (tabla del tamao dado).
    * @param tamano  tamao de la tabla de dispersin 
    * @exception ExcepcionParametroIncorrecto  si el tamao es negativo
    * @pre tamano>=0, new ExcepcionParametroIncorrecto("tamao negativo")
    */
   public ConjuntoTablaImpl(int tamano)
   {
      td = new TablaDispersion<E,E>(tamano);
   }

   
   /**
    * Accesor de lectura del nmero de elementos que hay al contenedor.
    * @return  nmero de elementos que contiene actualmente
    */
   public int numElems() { return  td.numElems(); }


   /**
    * Mtodo para comprobar si el contenedor est vaco.
    * @return  cierto o falso, segn si est vaco o no lo est
    */
   public boolean estaVacio() { return  td.estaVacio(); }

   
   /**
    * Aade un elemento, si se puede. Utiliza el mismo elemento como clave
    * y como valor.
    * @param elem  elemento que se quiere aadir al conjunto
    */
   public void insertar(E elem)
   {
   	td.insertar(elem, elem);
   }


   /**
    * Comprueba si hay un elemento con una determinada clave.
    * @param elem  elemento
    * @exception ExcepcionContenedorVacio  si la tabla est vaca
    * @exception ExcepcioElementsNoComparables  si el elemento es no
    *            comparable
    * @return  cierto o falso, segn si encuentra o no encuentra la clave
    */
   public boolean esta(E elem)
   {
      return  td.consultar(elem) != null;
   }


   /**
    * Borra la clave y el elemento asociado, si se puede.
    * @see  ClaveValor
    * @param elem elemento a borrar
    * @exception ExcepcionContenedorVacio  si la tabla est vaca
    * @exception ExcepcioElementsNoComparables  si el elemento es no
    *            comparable
    * @return  elemento borrado asociado con la clave;
    *          o null, si no estaba
    */
   public E borrar(E elem)
   {
   	return  td.borrar(elem);
   }


   /**
    * Accesor de lectura de los elementos que hay en el contenedor.
    * Retorna una enumeracin.
    * Enumerar es simplemente enunciar la una detrs la otra (las cosas
    * de una serie, las partes de un todo). Pero si el contenedor tiene definido
    * algn tipo de ordenacin o de recorrido, la enumeracin debe ser
    * consecuente y ofrecer los elementos por orden (FIFO, LIFO, inordre,
    * etc.), sin alterar el estado actual del contenedor.
    * @see  Iterador#haySiguiente()
    * @see  Iterador#siguiente()
    * @exception ExcepcionPosicionInvalida  si se quiere obtener el siguiente
    *            elemento de la enumeracin y ste no existe
    * @return  enumeracin de los elementos asociados con las claves
    */
   public Iterador<E> elementos() { return  td.elementos(); }


   /**
    * Mtodo que sobrescribe Object.toString().
    * Separa un par del siguiente con el salto de lnea de la
    * plataforma.
    */
   public String toString()
   {
	  return Utilidades.toStringContenedorDelegacion("ConjuntoTablaImpl",td);
   }
}
