package uoc.ei.tads;
/**
 * Clase que implementa las operaciones de un diccionario por delegacin en
 * un arbol binario AVL.
 *
 * Los diccionarios son estructuras que almacenan elementos con una clave
 * asociada. La clave debe disponer de una operacin de igualdad. En el caso
 * de los diccionarios ordenados debe permitir la ordenacin total entre claves.
 * Algunas implementaciones permiten claves repetidas y otras no. El elemento
 * asociado a la clave puede ser cualquier objeto.
 *
 * En esta implementacin se utilizan elementos de la clase ClaveValor,
 * que empareja la clave y el valor asociado. Si no hay comparador
 * especfico se utilitza ClaveValorComparable que implementa la interfaz
 * java.lang.Comparable, por delegacin en la clave.
 *
 * @see  ClaveValor
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informacin,
 *          Universitat Oberta de Catalunya (UOC)
 * @version 2.0.0
 */
public class DiccionarioAVLImpl<C,E>  implements Diccionario<C,E>
{
   /** Atributo que en Java 1.5 determina la compatibilidad entre objetos
    * serialitzables de la misma clase. El identificador se calcula
    * mediante un mtodo de la clase Utilidades. */
   private static final long serialVersionUID = Utilidades.getSerialVersionUID();
	
	
   /** Arbol binario AVL que implementa las operaciones por delegacin. */
   protected ArbolAVL<ClaveValor<C,E>> avl;

   /**
    * Constructor sin parmetros. Se espera que las clases de los elementos
    * implementen la interfaz java.lang.Comparable.
    */
   public DiccionarioAVLImpl() { avl = new ArbolAVL<ClaveValor<C,E>>(); }

   /**
    * Constructor con un parmetro y elementos de una clase comparable con
    * el comparador dado.
    * @param comparador  comparador que permite deducir la prioridad
    * @exception ExcepcionParametroIncorrecto  si el comparador es null
    */
   public DiccionarioAVLImpl(java.util.Comparator<ClaveValor<C,E>> comparador)
          throws  ExcepcionParametroIncorrecto
   {
      avl = new ArbolAVL<ClaveValor<C,E>>(comparador);
   }

   /**
    * Accesor de lectura del nmero de elementos que hay en el contenedor.
    * @return  nmero de elementos que contiene actualmente
    */
   public int numElems() { return  avl.numElems(); }

   /**
    * Mtodo para comprobar si el contenedor est vaco.
    * @return  cierto o falso, segn si est vaco o no lo est
    */
   public boolean estaVacio() { return  avl.estaVacio(); }

   /**
    * Aade un elemento con una clave asociada, si se puede. Si encuentra un
    * elemento con la misma clave lo sobrescribe.
    * @see  ClaveValor
    * @param clave  clave asociada al elemento que se quiere aadir
    * @param elem  elemento que se quiere aadir al diccionario
    */
   public void insertar(C clave,E elem)
   {
      avl.insertar( new ClaveValor<C,E>(clave, elem));
   }

   /**
    * Comprueba si hay un elemento con una determinada clave.
    * @param clave  clave asociada a un elemento
    * @return  cierto o falso, segn si encuentra o no encuentra la clave
    */
   public boolean esta(C clave)
   {
      return  consultar(clave) != null;
   }

   /**
    * Accesor de lectura del elemento asociado con una clave.
    * @see  ClaveValor
    * @param clave  clave de referencia
    * @return  elemento asociado con la clave; o null, si no era
    */
   public E consultar(C clave)
   {
   	E result=null;
   	ClaveValor<C,E> aux=new ClaveValor<C,E>(clave,null);
    ClaveValor<C,E> claveValor =  avl.consultar(aux);
    if (claveValor!=null)
    	result=claveValor.getValor();
    return result;
   }

   /**
    * Borra la primera clave coincidente y el elemento asociado, si se puede.
    * @see  ClaveValor
    * @param clave  clave de referencia
    * @return  elemento borrado asociado con la clave;
    *          o null, si no era
    */
   public E borrar(C clave)
   {
   	E result=null;
   	ClaveValor<C,E> aux=new ClaveValor<C,E>(clave,null);
    ClaveValor<C,E> claveValor =  avl.borrar(aux);
    if (claveValor!=null)
    	result=claveValor.getValor();
    return result;
   }

   /**
    * Accesor de lectura de los elementos que hay en el contenedor.
    * Retorna una enumeracin.
    * Enumerar es simplemente enunciar la una detrs la otra (las cosas
    * de una serie, las partes de un todo). Pero si el contenedor tiene definido
    * algn tipo de ordenacin o de recorrido, la enumeracin debe ser
    * consecuente y ofrecer los elementos por orden (FIFO, LIFO, inordre,
    * etc.), sin alterar el estado actual del contenedor.
    * @see  Iterador#haySiguiente()
    * @see  Iterador#siguiente()
    * @exception ExcepcionPosicionInvalida  si se quiere obtener el siguiente
    *            elemento de la enumeracin y ste no existe
    * @return  enumeracin de las claves del contenedor ordenadas
    */
   public Iterador<C> claves()
   {
      return  new IteradorRecorridoClavesImpl<C,E>(avl.recorridoInorden());
   }

   /**
    * Accesor de lectura de los elementos que hay en el contenedor.
    * Retorna una enumeracin.
    * Enumerar es simplemente enunciar la una detrs la otra (las cosas
    * de una serie, las partes de un todo). Pero si el contenedor tiene definido
    * algn tipo de ordenacin o de recorrido, la enumeracin debe ser
    * consecuente y ofrecer los elementos por orden (FIFO, LIFO, inordre,
    * etc.), sin alterar el estado actual del contenedor.
    * @see  Iterador#haySiguiente()
    * @see  Iterador#siguiente()
    * @exception ExcepcionPosicionInvalida  si se quiere obtener el siguiente
    *            elemento de la enumeracin y no hay ningn o no hay ningn
    *            ms
    * @return  enumeracin de los elementos asociados con las claves
    */
   public Iterador<E> elementos()
   {
      return  new IteradorRecorridoValoresImpl<C,E>(avl.recorridoInorden());
   }

   /**
    * Mtodo que sobrescribe Object.toString().
    * Separa una par de la siguiente con el salto de lnea de la
    * plataforma.
    */
   public String toString() {
	   return  Utilidades.toStringContenedorDelegacion("DiccionarioAVLImpl",avl);
   }
   
}
