package uoc.ei.tads.grafos;

/**
 * Clase que representa la arista de un grafo (dirigido o no dirigido).
 *
 * @author  Jordi Alvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informacin,
 *          Universitat Oberta de Catalunya (UOC)
 * @version 2.0.0
 */
abstract class AristaAbstracta<TEtiqueta,TElem> implements Arista<TEtiqueta,TElem>
{
   /** Primer vrtice de la arista. En el caso de una arista dirigida, representa
    * el origen. */
   protected Vertice<TElem> verticeA;

   /** Segundo vrtice de la arista. En el caso de una arista dirigida representa
    * el destino. */
   protected Vertice<TElem> verticeB;

   /** Valor de la etiqueta de la arista, o null si la relacin es no valorada. */
   protected TEtiqueta etiqueta;


   /**
    * Constructor con dos parmetros. No se permiten bucles.
    * @param vertice1  primer vrtice de la arista (u origen en el caso de dirigida)
    * @param vertice2  segundo vrtice de la arista (o destino en el caso de dirigida)
    * @pre vertice1!=null && vertice2!=null, ExcepcionParametroIncorrecto
    */
   public AristaAbstracta(Vertice<TElem> vertice1, Vertice<TElem> vertice2)
   {
      this.verticeA = vertice1;
      this.verticeB = vertice2;
   }


   /**
    * Accesor de lectura del valor asociado a la arista.
    * @return  valor de la arista; o null, si es una relacin no valuada
    */
   public TEtiqueta getEtiqueta() { return etiqueta; }

   
   /**
    * Accesor de escritura del valor asociado a la arista.
    * @pre true
    * @param valor  valor de la relacin
    * @post getValor()@pre == valor
    */
   public void setEtiqueta(TEtiqueta etiqueta) { this.etiqueta = etiqueta; }
  

   /**
    * Indica si la arista incide en un vrtice dado, como origen o
    * destino.
    * @param v  vrtice del cual se quiere saber si pertenece a la arista
    * @pre v != null
    * @post true
    * @return  cierto si la arista incide en el vrtice; falso en otro caso.
    * @pre v!=null, ExcepcionParametroIncorrecto
    */
   public boolean incideEnVertice(Vertice<TElem> vertice)
   {
      return ( this.verticeA.equals(vertice) || this.verticeB.equals(vertice) );
   }


   /**
    * Mtodo que sobrescribe Object.equals(Object obj). Delega en los
    * mismos mtodos implementados o heredados por los vrtices.
    * @return  cierto si el objeto recibido como parmetro es de la clase
    *          AristaDirigida y tiene el mismos vrtices en el mismo orden;
    *          sino retorna falso
    */
   public boolean equals(Object object)
   {
   	  if (getClass()!=object.getClass())
   	  	return false;
   	  AristaAbstracta<TEtiqueta,TElem> aresta=(AristaAbstracta<TEtiqueta,TElem>)object;
   	  return verticeA.equals(aresta.verticeA) &&
   	  		 verticeB.equals(aresta.verticeB);
   }


   /**
    * Mtodo que sobrescribe Object.toString(). Delega en los mismos
    * mtodos implementados o heredados por los vrtices y por la etiqueta (si
    * hay).
    * @return  cadena de caracteres con el primer vrtice, la etiqueta si
    *          tiene y el segundo vrtice
    */
   public String toString()
   {
   	  return "("+verticeA+":"+etiqueta+":"+verticeB+")";
   }
}
